import java.util.*;
import Euler;

/**
N3 parser.
  @version 27.035
  @author Jos De Roo
                                                                                                    */
public class Parser {
  StringTokenizer st = null;                // raw N3 string tokenizer
  Vector vt = null;                         // quantified variables table
  Euler r = null;                           // root Euler object
  String u = null;                          // base URI of the RDF resource

  static final String version = "27.035";
  static final String ST = "{}[]()<>\";,. \t\r\n\\";
  static final String N3a = "a";
  static final String N3e = "=";
  static final String RDFtype = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>";
  static final String DAMLequivalentTo = "<http://www.daml.org/2001/03/daml-ont#equivalentTo>";
  static final String DAMLUniqueProperty = "<http://www.daml.org/2001/03/daml-ont#UniqueProperty>";
  static final String ENonIrreflexiveProperty = "<http://www.agfa.com/w3c/euler#NonIrreflexiveProperty>";

/**
  constructs an N3 parser
  @param s the N3 string
  @param vartab universally quantified variables table
  @param root Euler object
  @param base URI of the RDF resource
                                                                                                    */
  public Parser(String s, Vector vartab, Euler root, String uri) {
    st = new StringTokenizer(s, ST, true);
    vt = vartab;
    r = root;
    u = uri;
  }

/**
  N3 triple parse method
  @return Euler object
                                                                                                    */
  public Euler parse() {
    String nt = tokenize(true);
    if (nt == null) return null;
    return parse(true, nt);
  }

/**
  N3 node parse method
  @param nt next token
  @return Euler object
                                                                                                    */
  public Euler parse(String nt) {
    return parse(false, nt);
  }

  Euler parse(boolean b, String nt) {
    Euler e = new Euler();
    try {
      if (b) {
        if (nt == null) e.subject = null;
        else if (nt.equals("{")) {
          nt = tokenize(true);
          e.subject = parse(true, nt);
          Euler el = e.subject;
          nt = tokenize(true);
          while (!nt.equals("}")) {
            el.near = parse(true, nt);
            el = el.near;
            nt = tokenize(true);
          }
        }
        else if (nt.equals("[")) {
          e.subject = parse(true, null);
          r.anon.put(e.subject.toString(), e.subject);
          nt = tokenize(true);
          if (!nt.equals("]")) System.err.println("** (p1) expecting ] at " + e + " but got " + nt);
        }
        else e.subject = parse(false, nt);
        nt = tokenize(true);
      }
      if (nt.equals("[")) {
        String a = "ap" + vt.size();
        vt.addElement(a);
        Euler ap = parse(true, a);
        e.cverb = a;
        nt = tokenize(true);
        if (!nt.equals("]")) System.err.println("** (p2) expecting ] at " + e + " but got " + nt);
        e.near = ap;
      }
      else e.cverb = nt;
      e.verb = e.cverb;
      if (e.cverb.length() > 0 && !e.cverb.equals(N3a) && !e.cverb.equals(N3e)
          && e.cverb.charAt(0) != '(' && e.cverb.charAt(0) != '<'
          && e.cverb.charAt(0) != '[' && e.cverb.charAt(0) != '_'
          && e.cverb.charAt(0) != '"' && e.cverb.indexOf(':') != -1 ) {
        String pf = e.cverb.substring(0, e.cverb.indexOf(':') + 1);
        String pg = (String)r.nsp.get(pf);
        if (pg != null) pf = pg;
        String qf = (String)r.ns.get(pf);
        if (qf == null) {
          System.err.println("** no @prefix " + pf + " found, taking <" + u + "#>");
          qf = "<" + u + "#>";
          r.ns.put(pf, qf);
        }
        StringBuffer sb = new StringBuffer(pf);
        sb.append(e.cverb.substring(e.cverb.indexOf(':') + 1));
        e.cverb = sb.toString();
        sb = new StringBuffer(qf);
        sb.insert(sb.length() - 1, e.cverb.substring(e.cverb.indexOf(':') + 1));
        e.verb = sb.toString();
      }
      else if (e.cverb.length() > 0 && e.cverb.charAt(0) == '<' && e.cverb.indexOf(':') == -1) {
        StringBuffer sb = new StringBuffer(e.cverb);
        if (e.cverb.charAt(1) == '#') sb.insert(1, u);
        else sb.insert(1, u.substring(0, u.lastIndexOf('/') + 1));
        e.verb = sb.toString();
      }
      e.varid = vt.indexOf(e.verb);
      if (e.varid == -1) e.bound = true;
      if (nt.equals(".")) {
        e.object = parse(false, "");
        e.verb = "";
        e.cverb = "";
        e.bound = true;
        return e;
      }
      if (b) {
        nt = tokenize(true);
        if (nt.equals("{")) {
          nt = tokenize(true);
          e.object = parse(true, nt);
          Euler el = e.object;
          nt = tokenize(true);
          while (nt != null && !nt.equals("}")) {
            el.near = parse(true, nt);
            el = el.near;
            nt = tokenize(true);
          }
        }
        else if (nt.equals("[")) {
          e.object = parse(true, null);
          r.anon.put(e.object.toString(), e.object);
          nt = tokenize(true);
          if (!nt.equals("]")) System.err.println("** (p3) expecting ] at " + e + " but got " + nt);
        }
        else {
          e.object = parse(false, nt);
          if (e.verb.equals(N3a) || e.verb.equals(RDFtype)) {
            e.verb = e.object.verb;
            e.varid = e.object.varid;
            e.bound = e.object.bound;
            if (e.verb.equals(ENonIrreflexiveProperty)) r.refl.addElement(e.subject.verb);
            else if (e.verb.equals(DAMLUniqueProperty)) r.uniq.addElement(e.subject.verb);
          }
        }
        nt = tokenize(true);
        if (nt.endsWith("\"\"\"")) {
          e.object.cverb = e.object.cverb + nt;
          e.object.verb = e.object.cverb;
          nt = tokenize(true);
        }
        if (e.verb.equals(N3e)) synon(e, e, r);
        syn(e, r);
        Euler el = e;
        while (nt.equals(";") || nt.equals(",")) {
          while (el.near != null) el = el.near;
          if (nt.equals(";")) {
            nt = tokenize(true);
            if (nt.equals("]")) {
              r.verb = ";]";
              nt = ";";
              break;
            }
            else if (nt.equals(".")) break;
          }
          else nt = el.cverb;
          el.near = parse(false, nt);
          el.near.subject = e.subject;
          nt = tokenize(true);
          if (nt.equals("{")) {
            nt = tokenize(true);
            el.near.object = parse(true, nt);
            Euler ef = el.near.object;
            nt = tokenize(true);
            while (nt != null && !nt.equals("}")) {
              ef.near = parse(true, nt);
              ef = ef.near;
              nt = tokenize(true);
            }
          }
          else if (nt.equals("[")) {
            el.near.object = parse(true, null);
            r.anon.put(el.near.object.toString(), el.near.object);
            nt = tokenize(true);
            if (!nt.equals("]")) System.err.println("** (p4) expecting ] at " + e + " but got " + nt);
          }
          else {
            el.near.object = parse(false, nt);
            if (el.near.verb.equals(N3a) || el.near.verb.equals(RDFtype)) {
              el.near.verb = el.near.object.verb;
              el.near.varid = el.near.object.varid;
              el.near.bound = el.near.object.bound;
              if (el.near.verb.equals(ENonIrreflexiveProperty)) r.refl.addElement(el.near.subject.verb);
              else if (el.near.verb.equals(DAMLUniqueProperty)) r.uniq.addElement(el.near.subject.verb);
            }
          }
          nt = tokenize(true);
          if (nt.endsWith("\"\"\"")) {
            el.near.object.cverb = el.near.object.cverb + nt;
            el.near.object.verb = el.near.object.cverb;
            nt = tokenize(true);
          }
          if (el.near.verb.equals(N3e)) {
            synon(e, el.near, r);
            el.near = null;
          }
          syn(el.near, r);
        }
        if (!nt.equals(".") && !nt.equals(";"))
          System.err.println("** (p5) expecting . or ; at " + e + " but got " + nt);
      }
      e.far = r;
      return e;
    }
    catch (NullPointerException exc) {
      System.err.println("** @@ trouble with " + e);
      return e;
    }
  }

  String tokenize(boolean ss) {
    String nt = token(ss);
    while (st.hasMoreTokens() && nt == null) nt = token(ss);
    return nt;
  }

  String token(boolean ss) {
    if (r.verb != null && r.verb.equals(".}")) return r.verb = "}";
    if (r.verb != null && r.verb.equals(";]")) return r.verb = "]";
    if (!st.hasMoreTokens()) return null;
    String t = null;
    String nt = st.nextToken();
    if (ss && nt.indexOf('#') != -1) {
      t = nt.substring(0, nt.indexOf('#'));
      String v = r.verb;
      while (nt != null && !nt.equals("\n")) nt = tokenize(false);
      r.verb = v;
      if (t.equals("")) return null;
    }
    else if (ss && nt.equals("@prefix") || nt.equals("bind")) {
      String nsc = tokenize(ss);
      if (nsc.equals("default")) nsc = ":";
      String nsd = nsc;
      String nsu = tokenize(ss);
      StringBuffer sb = new StringBuffer(nsu);
      if (nsu.equals("<>")) sb.insert(sb.length() - 1, '#');
      if (sb.length() > 0 && nsu.indexOf(':') == -1) {
        if (sb.charAt(1) == '#') sb.insert(1, u);
        else if (sb.charAt(1) == '/') sb.insert(1, u.substring(0, u.lastIndexOf('/')));
        else sb.insert(1, u.substring(0, u.lastIndexOf('/') + 1));
      }
      nsu = sb.toString();
      String nsv = (String)r.ns.get(nsd);
      while (nsv != null && !nsu.equals(nsv)) {
        nsd = "ns" + nsd;
        nsv = (String)r.ns.get(nsd);
      }
      r.nsp.put(nsd, nsd);
      r.nsp.put(nsc, nsd);
      r.ns.put(nsd, nsu);
      tokenize(ss);
      return null;
    }
    else if (nt.equals("\"")) {
      StringBuffer sb = new StringBuffer("\"");
      if (r.verb.equals("\"\"")) {
        while (true) {
          sb.append(st.nextToken());
          if (sb.toString().endsWith("\"\"\"")) break;
        }
        t = sb.toString();
      }
      else {
        nt = st.nextToken();
        while (!nt.equals("\"")) {
          if (nt.equals("\\")) {
            sb.append(nt);
            nt = st.nextToken();
          }
          sb.append(nt);
          if (!st.hasMoreTokens()) {
            System.err.println("** (t1) expecting \" at " + sb);
            break;
          }
          nt = st.nextToken();
        }
        t = sb.append("\"").toString();
      }
    }
    else if (ss && nt.equals("<")) {
      StringBuffer sb = new StringBuffer("<");
      nt = st.nextToken();
      while (!nt.equals(">")) {
        sb.append(nt);
        if (!st.hasMoreTokens()) {
          System.err.println("** (t2) expecting > at " + sb);
          break;
        }
        nt = st.nextToken();
      }
      t = sb.append(">").toString();
    }
    else if (ss && nt.equals("(")) {
      StringBuffer sb = new StringBuffer("(");
      nt = tokenize(ss);
      while (!nt.equals(")")) {
        if (sb.charAt(sb.length() - 1) != '(') sb.append(' ');
        sb.append(nt);
        if (!st.hasMoreTokens()) {
          System.err.println("** (t3) expecting ) at " + sb);
          break;
        }
        nt = tokenize(ss);
      }
      t = sb.append(")").toString();
    }
    else if (ss && (nt.equals(" ") || nt.equals("\t") || nt.equals("\r") || nt.equals("\n")
      || nt.equals("has") || nt.equals("is") || nt.equals("of") || nt.equals("-") || nt.equals(">")))
      return null;
    else t = nt;
    if (r.verb != null && t != null && !r.verb.equals(".") && t.equals("}")) {
      r.verb = ".}";
      return ".";
    }
    if (r.verb != null && t != null && !r.verb.equals(";") && t.equals("]")) {
      r.verb = ";]";
      return ";";
    }
    return r.verb = t;
  }

  void synon(Euler e, Euler t, Euler r) {
    if (t.subject != null && t.object != null && t.subject.subject == null && t.subject.object == null
      && t.object.subject == null && t.object.object == null) r.syn.put(t.subject.verb, t.object.verb);
    else if (t.subject != null && t.object == null
      && t.subject.subject == null && t.subject.object == null) r.syn.put(t.subject.verb, e);
    else if (t.subject == null && t.object != null
      && t.object.subject == null && t.object.object == null) r.syn.put(t.object.verb, e);
  }

  void syn(Euler t, Euler r) {
    if (t == null) return;
    if (t.subject != null && t.subject.verb != null) {
      if (t.subject.subject == null && t.subject.object == null) {
        Object o = r.syn.get(t.subject.verb);
        if (o != null) {
          if (o instanceof String) t.subject.verb = (String)o;
          else t.subject = ((Euler)o).eval(null);
        }
      }
    }
    if (t.verb != null) {
      Object o = r.syn.get(t.verb);
      if (o != null) {
        if (o instanceof String) t.verb = (String)o;
        else t.verb = o.toString();
      }
    }
    if (t.object != null && t.object.verb != null) {
      if (t.object.subject == null && t.object.object == null) {
        Object o = r.syn.get(t.object.verb);
        if (o != null) {
          if (o instanceof String) t.object.verb = (String)o;
          else t.object = ((Euler)o).eval(null);
        }
      }
    }
  }
}
